#include <Wire.h>
#include "MCP23017.h"
#include "WiFly.h"
#include "pitches.h"
#include "FreqOut.h"

// Twitter
#define LIB_DOMAIN "arduino-tweet.appspot.com"
#define TWITTER_OAUTH_TOKEN "<your-oAuth-token-here>"

// pins
#define SPEAKER_PIN 9
#define HANDSET_PIN 2
#define DIALER_PIN 3

// handset pin states
#define HANDSET_DOWN HIGH
#define HANDSET_UP LOW

// Twitter related
#define HAPPY ":~)"
#define  SAD ":~("
#define  SHOCKED ":-o"
#define  EXHAUSTED "#-:)"
#define  CRYING ":'-("
#define  DRUNK "%-)"
#define  ENJOYING_THE_SUN "0|~)"
#define  ANGRY "X-("
#define  BAD_HAIR_DAY "?:~)"
#define  BROKEN_HEARTED "</3"

char* TWEETS [] = {
  HAPPY,            // 1
  SAD,              // 2
  SHOCKED,          // 3
  EXHAUSTED,        // 4
  CRYING,           // 5
  DRUNK,            // 6
  ENJOYING_THE_SUN, // 7
  ANGRY,            // 8
  BAD_HAIR_DAY,     // 9
  BROKEN_HEARTED    // 0
};
boolean wifiEnabled;
Client client(LIB_DOMAIN, 80);
char msg[4+1]; //1 for null termination
char* msgPtr;


// Color Management
MCP23017 portExpander;
boolean digits[10];
int stepper = 0;
int absStepper = 0;
int currDigit = 1;
boolean currState = false;

// LED Management
byte bit0state = 0;
byte bit1state = 0;
byte bit9state = 0;
byte byte0, byte1;

// Handset & Dialing Management
enum {
  STATE_ASSOCIATING_WIFLY,
  STATE_ON_HOOK,
  STATE_OFF_HOOK,
  STATE_DIALING,
  STATE_DIALING_TIMEOUT_ERROR,
  STATE_NUMBER_DIALED,  
  STATE_CONNECTING,
  STATE_CONNECTING_ERROR,  
  STATE_SENDING,
  STATE_AWAITING_REPLY,
  STATE_REPLY_TIMEOUT_ERROR,
  STATE_ERROR,
  STATE_NO_LINE_ERROR,  
  STATE_TWEET_SENT,
  STATE_SENDING_ERROR,
  STATE_SILENCE,
} PhoneState;

// INITIAL STATE
int state;
boolean handsetState;

boolean dialerState = true;
long lastDialerToggle = 0;
int dialedNumberAcc = 0;
int awaitingReplyTimeout;
boolean awaitingReplySoundOn;
int awaitingReplyCount;
long startDialingTStamp;
long silenceStartedTStamp;

void setup() {
  Serial.begin(115200);

  state = STATE_ASSOCIATING_WIFLY; 
  handsetState = HANDSET_DOWN;
  
  pinMode(HANDSET_PIN, INPUT);
  digitalWrite(HANDSET_PIN, HIGH); // turn on pull-up
  pinMode(DIALER_PIN, INPUT);
  digitalWrite(DIALER_PIN, HIGH); // turn on pull-up
  pinMode(SPEAKER_PIN, OUTPUT);
  Wire.begin();
  portExpander.SetDir(0,0x00);  // all output
  portExpander.SetDir(1,0x00);  // all output
  setAllLeds(false);

  cheesy_effect();
}

void loop() {
  switch(state) {
    case STATE_ASSOCIATING_WIFLY:       state = handleAssociatingWifly();      break;
    case STATE_ON_HOOK:                 state = handleOnHook();                break;
    case STATE_OFF_HOOK:                state = handleOffHook();               break;
    case STATE_DIALING:                 state = handleDialing();               break;
    case STATE_DIALING_TIMEOUT_ERROR:   state = handleDialingTimeout();        break;
    case STATE_NUMBER_DIALED:           state = handleNumberDialed();          break;
    case STATE_CONNECTING:              state = handleConnecting();            break;
    case STATE_CONNECTING_ERROR:        state = handleConnectingError();       break;
    case STATE_SENDING:                 state = handleSending();               break;
    case STATE_AWAITING_REPLY:          state = handleAwaitingReply();         break;
    case STATE_REPLY_TIMEOUT_ERROR:     state = handleReplyTimeoutError();     break;
    case STATE_NO_LINE_ERROR:           state = handleNoLineError();           break;
    case STATE_TWEET_SENT:              state = handleTweetSent();             break;
    case STATE_SENDING_ERROR:           state = handleSendingError();          break;
    case STATE_SILENCE:                 state = handleSilence();               break;
  }
  if (state != STATE_ON_HOOK) {
    if(!isHandsetOffHook()) {
      setAllLeds(false);
      noTone(SPEAKER_PIN);
      state = STATE_ON_HOOK;

      // todo: add reset behavior per start-state?
      // TODO: reset client state if in the middle of a network operation
    }
  }
}

int handleOffHook() {
  // name says it all, doesn't it?
  cheesy_color_effect(true); 
  // cleanup
  dialedNumberAcc = 0;
  startDialingTStamp = millis();
  tone(SPEAKER_PIN, NOTE_G4);
  return STATE_DIALING;
}

int handleOnHook() {
  noTone(SPEAKER_PIN);
  return isHandsetOffHook()? STATE_OFF_HOOK : STATE_ON_HOOK;
}

int handleAssociatingWifly() {
  Serial.println("Wifly, begin!");
  WiFly.begin();
  blinkAll(3);
  
  if(tryConnectWiFly()) {
    handleWiFlySuccess(3);
  }
  else {
    handleWiFlyFailure(3);
  }
  return STATE_ON_HOOK;
}

boolean tryConnectWiFly() {
  Serial.println("now, join network!");
  return WiFly.join("PNINUSH");
}

boolean isHandsetOffHook() {
  return (safeDigitalRead(HANDSET_PIN) == HANDSET_UP);
}

int handleDialing() {
  if((millis() - startDialingTStamp) > 10000) {
    setAllLeds(false);
    return STATE_DIALING_TIMEOUT_ERROR;
  }
    
  long timeSinceLastDialerToggle = micros() - lastDialerToggle;
  if(timeSinceLastDialerToggle > 100000) {
    if(dialedNumberAcc > 0) {
      return STATE_NUMBER_DIALED;
    }
  }
  boolean b = safeDigitalRead(DIALER_PIN);
  if(b != dialerState) {
    noTone(SPEAKER_PIN);
    dialerState = b;
    if(b == LOW) {
      handleDialerLowEvent(); 
    }
    else {
      handleDialerHighEvent();
    }
  }
  return STATE_DIALING;
}

int handleDialingTimeout() {
  flipAllLeds();  
  for(int i = 0; i < 10; ++i) {
    tone(SPEAKER_PIN, NOTE_A4);
    delayWhileUpdatingLeds(10);
    noTone(SPEAKER_PIN);
    delayWhileUpdatingLeds(10);
  }
  flipAllLeds();
  delayWhileUpdatingLeds(200);
  return STATE_DIALING_TIMEOUT_ERROR;
}

int handleNumberDialed() {
  noTone(SPEAKER_PIN);
  Serial.print("Number dialed: ");Serial.println(dialedNumberAcc);
  if(!wifiEnabled) {
    Serial.println("trying to reconnect...");
    if(tryConnectWiFly()) {
      handleWiFlySuccess(2);
    }
    else {
      handleWiFlyFailure(2);
    }
  }
  if(wifiEnabled) {
    return STATE_CONNECTING;
  }
  else {
    delay(1000); // for better user experience?
    return STATE_NO_LINE_ERROR;
  }
}


int handleConnecting() {
  msgPtr = TWEETS[dialedNumberAcc - 1];
  Serial.print("tweeting: "); Serial.println(msgPtr);
  Serial.println("connecting...");
  // TODO: Make connection Noise
  if (!client.connect()) {
    Serial.println("connection failed");
    return STATE_CONNECTING_ERROR;
  }
  return STATE_SENDING;
}

int handleSending() {
  byte size = strlen(msgPtr);
  Serial.print("size: "); Serial.println((int)size);
  // TODO: make some noise
  for(int i = 0; i < strlen(msgPtr); ++i) {
    msg[i] = msgPtr[i];
  }
  msg[size]=0; // null termination
  long time = millis();
  int digitCount = getDigitCount(time);

  Serial.println("connected");
  client.println("POST http://" LIB_DOMAIN "/update HTTP/1.0");
  client.print("Content-Length: ");
  client.println(size  + strlen(TWITTER_OAUTH_TOKEN) + 14 + digitCount + 3);
  client.println();
  client.print("token=");
  client.print(TWITTER_OAUTH_TOKEN);
  client.print("&status=");
  client.print(msg);
  client.print(" ("); client.print(time); client.println(")\"");
  Serial.print("time: "); Serial.println(time);
  Serial.print("digitCount: "); Serial.println(digitCount);
  
  client.flush();
  Serial.println("sent!");
  awaitingReplyTimeout = 20000;
  awaitingReplySoundOn = false;
  awaitingReplyCount = 0;
  return STATE_AWAITING_REPLY;
}

int getDigitCount(long number) {
  int result = 0;
  while(number > 0) {
    result++;
    number /= 10;
  }
  return result; 
}


int handleAwaitingReply() {
  boolean succeeded = false;
  if(awaitingReplyTimeout <= 0) {
    return STATE_REPLY_TIMEOUT_ERROR;
  }

  if (client.available()) {
    succeeded = true; // assuming getting something from server is succeeding   
    char c = client.read();
    Serial.print(c);
  }
  awaitingReplyCount++;
  int currDelay = awaitingReplySoundOn ? 400 : 1200;
  if (awaitingReplyCount >= currDelay) {
    awaitingReplyCount = 0;
    if (awaitingReplySoundOn) {
      noTone(SPEAKER_PIN);
      awaitingReplySoundOn = false;
    }
    else {
      tone(SPEAKER_PIN, NOTE_A3);
      awaitingReplySoundOn = true;
    }
  }
  
  if (!client.connected()) {
    Serial.println();
    Serial.println("disconnecting.");
    client.stop();
    noTone(SPEAKER_PIN);
    if(!succeeded) {
      return STATE_SENDING_ERROR;
    }
    return STATE_TWEET_SENT;
  }
  awaitingReplyTimeout--;
  delayMicroseconds(1000);
  return STATE_AWAITING_REPLY;
}

int handleSendingError() {
  //TODO: add timeout for silence
  tone(SPEAKER_PIN, NOTE_C3);
  delay(200);
  noTone(SPEAKER_PIN);
  delay(200);
  return STATE_SENDING_ERROR;
}

int handleTweetSent() {
  Serial.println("Tweet sent!");
  setAllLeds(true);
  toneWithLight(NOTE_C4, 100);
  toneWithLight(NOTE_E4, 100);
  toneWithLight(NOTE_G4, 100);
  toneWithLight(NOTE_C5, 100);
  toneWithLight(0, 200);
  toneWithLight(NOTE_C5, 100);
  toneWithLight(0, 5);
  toneWithLight(NOTE_C5, 400);
  toneWithLight(0, 5);
  silenceStartedTStamp = millis();
  return STATE_SILENCE;
}

void toneWithLight(int freq, int duration) {
  flipAllLeds();
  if(freq) {
    tone(SPEAKER_PIN, freq);
  }
  else {
    noTone(SPEAKER_PIN);
  }
  delayWhileUpdatingLeds(duration);
}


int handleSilence() {
  if(millis() - silenceStartedTStamp > 10000) {
     return STATE_DIALING_TIMEOUT_ERROR;
  }
  return STATE_SILENCE;
}

int handleReplyTimeoutError() {
  //TODO: add timeout for silence
  tone(SPEAKER_PIN, NOTE_C4);
  delay(200);
  noTone(SPEAKER_PIN);
  delay(200);
  return STATE_REPLY_TIMEOUT_ERROR;
}

int handleNoLineError() {
  //TODO: add timeout for silence
  tone(SPEAKER_PIN, NOTE_C5);
  delay(200);
  noTone(SPEAKER_PIN);
  delay(200);
  return STATE_NO_LINE_ERROR;
}

int handleConnectingError() {
  //TODO: add timeout for silence
  tone(SPEAKER_PIN, NOTE_C6);
  delay(200);
  noTone(SPEAKER_PIN);
  delay(200);
  return STATE_CONNECTING_ERROR;
}

///////////////////////////////////
// Event Handlers

void handleDialerHighEvent() {
  // tone(SPEAKER_PIN, NOTE_E4);
  lastDialerToggle = micros();
}

void handleDialerLowEvent() {
  // noTone(SPEAKER_PIN);
  long now = micros();
  if(lastDialerToggle != 0) {
    long delta = now - lastDialerToggle;
    // Serial.print("delta(LOW)=");Serial.println(delta / 1000);
    if(delta > 55000 && delta < 65000) {
      dialedNumberAcc++;
    }
  }
  lastDialerToggle = now;
}


void handleWiFlyFailure(int alertLevel) {
  wifiEnabled = false;

  if(alertLevel > 0) {
    Serial.println("Wifi association failed.");  
  }
  if(alertLevel > 1) { 
    tone(SPEAKER_PIN, NOTE_C4);
    delay(100);
    tone(SPEAKER_PIN, NOTE_A3);
    delay(100);
    tone(SPEAKER_PIN, NOTE_A3);
    delay(100);
    noTone(SPEAKER_PIN);
    delay(100);
    tone(SPEAKER_PIN, NOTE_C4);
    delay(100);
    tone(SPEAKER_PIN, NOTE_A3);
    delay(100);
    tone(SPEAKER_PIN, NOTE_A3);
    delay(100);
    noTone(SPEAKER_PIN);
  }
  if(alertLevel > 2) {
    blinkAll(3);
  }
}

void handleWiFlySuccess(int alertLevel) {
  wifiEnabled = true;

  if(alertLevel > 0) {
    Serial.println("Wifi association succeeded!");
  }
  if(alertLevel > 1) { 
    tone(SPEAKER_PIN, NOTE_G3);
    delay(100);
    tone(SPEAKER_PIN, NOTE_A3);
    delay(100);
    tone(SPEAKER_PIN, NOTE_B3);
    delay(100);
    tone(SPEAKER_PIN, NOTE_C4);
    delay(100);
    tone(SPEAKER_PIN, NOTE_D4);
    delay(100);
    noTone(SPEAKER_PIN);
  }
  if(alertLevel > 2) {
    doTheNightRider(2);
  }
}


/////////////////////////////////////////////////////
// LED Management & Effects

void setAllLeds(boolean ledState) {
  for(int i = 0; i < 10; ++i) {
    digits[i] = ledState; 
  }
  updateLeds();
}

void flipAllLeds() {
  for(int i = 0; i < 10; ++i) {
    digits[i] = !digits[i]; 
  }
  updateLeds();
}

void blinkAll(int times) {
  for(int i = 0; i < times; ++i) {
    setAllLeds(true);
    delayWhileUpdatingLeds(300);
    setAllLeds(false);
    delayWhileUpdatingLeds(300);
  }
}

void doTheNightRider(int times) {
  for(int t = 0; t < times; ++t) {
    for(int i = 0; i < 10; ++i) {
      setAllLeds(false);
      digits[i] = true;
      digits[9-i] = true;
      for(int j = 0; j < 100; ++j) {
        updateLeds();
      }
    }
  }
  setAllLeds(false);
}

boolean cheesy_color_effect(boolean blocking) {
  if(blocking) {
    while(!cheesy_color_effect());
    return true;
  }
  else {
    return cheesy_color_effect();
  }
}

boolean cheesy_color_effect() {
  if(absStepper < 20) {
    if(stepper++ % 40 == 0) {
      if(currDigit == 1) {
        currState = !currState;
      }
      digits[currDigit] = currState;
      currDigit = (currDigit + 1) % 10;
      absStepper++;
    }
    updateLeds();
    return false;
  }
  else {
    absStepper = 0;
    return true;
  }
}

void delayWhileUpdatingLeds(int delayPeriod) {
  long start = micros();
  for(int i = 0; i <  delayPeriod; ++i) {
    updateLeds();
    delayMicroseconds(300);
  }
}

void updateLeds() {

  byte0 = 0; // regular leds feed on Vss
  byte1 = 255; // bi-color leds feed on ground, so 1=0 in their case

  if(digits[0]) {
    byte bit0 = bit0state ? 4 : 8;
    byte1 &= (255 - bit0);
    bit0state = !bit0state;
  }
  if(digits[1]) {
    byte bit1 = (bit1state%10 == 0) ? 16 : 32;
    byte1 &= (255 - bit1);
    bit1state++;
  }
  if(digits[9]) {
    byte bit9 = bit9state ? 1 : 2;
    byte1 &= (255 - bit9);
    bit9state = !bit9state;
  }
  for(int i = 2; i < 9; ++i) {
    if(digits[i]) {
      byte0 |= (1 << (i - 1));
    } 
  }
  portExpander.WriteByte(0, byte0);
  portExpander.WriteByte(1, byte1);
}


///////////////////////////////////
// Utility methods

boolean safeDigitalRead(byte pin) {
  boolean sample = digitalRead(pin);
  int consistentReadCount = 1;
  while(consistentReadCount < 3) {
    delayMicroseconds(500);
    if(digitalRead(pin) == sample) {
      consistentReadCount++;
    }
    else {
      sample = digitalRead(pin);
      consistentReadCount = 1;
    }
  }
  return sample;
}


float majScale[] = {
	A,  B,  CS,  D,  E,  FS,  GS,  A2_,   B2,  C2S,  D2,  E2,  F2S,  G2S,  A3_};

void cheesy_effect() { 
	int ptime;
	int   k,  x, dur, freq, t; 
	int i, j;
	float ps;         // variable for pow pitchShift routine
	float noteval;
	//note values
	
	for(i= 0; i<=4; i++){ 
		ps = (float)i / 12;         // choose new transpose interval every loop
                setAllLeds(true);
		for(x= 0; x<=15; x++){                      
			noteval = (majScale[x] / 64) * pow(2,ps);    // transpose scale up 12 tones - pow function generates transpostion
			dur = 50 - i * 8 - x;
			freqout(SPEAKER_PIN, (int)noteval, dur);
			delay(dur / 10);
                        if(x == 7) {
                          setAllLeds(false);
                        }
		}
	}
	
	for(i= 0; i<=6; i++){ 
		ps = (float)i / 12;         // choose new transpose interval every loop
                setAllLeds(true);
		for(x= 0; x<=15; x++){                      
			noteval = (majScale[x] / 64) * pow(2,ps);    // transpose scale up 12 tones - pow function generates transpostion
			dur = 5;
			freqout(SPEAKER_PIN, (int)noteval, dur);
			delay(dur / 10);
                        if(x == 7) {
                          setAllLeds(false);
                        }
		}
	}
} 

